freeze;

/**************************************************************************
 *                                                                        *
 * B. Eick, M. Kirschmer, C.R. Leedham-Green,                             *
 * The explicit membership problem for discrete free subgroups of rank 2  *
 * of SL(2,R)                                                             *
 *                                                                        *
 * Some supplementary Magma code, M. Kirschmer, February 2013             *
 *                                                                        *
 **************************************************************************/

declare type GrpSL2Gen;

declare attributes GrpSL2Gen:
  A, B,		// user given generators
  U, V,		// normalized witness pair
  WordsUV,	// how to express \bar{U},\bar{V} in {\bar{A},\bar{B}}
  S,		// how to conjugate U, V into "nice generators".
  K,		// base field for the original generators
  plc,		// embedding K -> R which yields G -> SL2(R)
  PSL,		// should we interpret G as a subgroup of PSL2(R)?
  finite,	// data for finite groups
  type;		
/* 
type can be one of the following
"nd" -> not discrete,
<0,0,0> -> free group & discrete
<a,0,0> -> Z/aZ * Z free product & discrete
<a,b,0> -> Z/aZ * Z/bZ free product & discrete
<a,b,c> -> Triangle group <a,b,c> & discrete
"ed" -> elementary & discrete
"finite" -> group finite

Currently only the cases finite, elementary and free are implemented
*/

myNF:= func< plc | Type(plc) eq Infty select Rationals() else NumberField(plc) >;
function myEval(x, plc : ForceReal:= false, Precision:= 300)
  if Type(plc) eq Infty then
    return ForceReal select RealField(Precision) ! x else x;
  end if;
  return Evaluate(x, plc : Precision:= Precision);
end function;
normalize:= func< A, plc | myEval(Trace(A), plc) lt 0 select -A else A >;
l2norm:= func< x | x[1]^2 + x[2]^2 >;
// Test if Q is in the circle centered at (Q1, 0) with radius r^(1/2)
InCircle:= func< P, Q1, r, plc | myEval(r - ((P[1] - Q1)^2 + P[2]^2) , plc) gt 0 >;

oo:= Infinity();
MAXTRIES:= 100;		
// If we read X^MAXTRIES in the constructive membership, we assume that we actually have X^n for some large n

procedure swap(~X, ~Y)
  temp:= X; X:= Y; Y:= temp;
end procedure;

function RationalReconstruction(x,y)
  d:= y-x;
  assert d gt 0;
  l:= Max(0, Ceiling(-Log(d/2)/Log(10))) + 2;
  r:= Round( (x+y)/2 * 10^l ) / 10^l;
  assert x lt r and r lt y;
  return r;
end function;

cross:= function(a,b,c,d)
try
  if a cmpeq oo then return (b-d)/(b-c); end if;
  if b cmpeq oo then return (a-c)/(a-d); end if;
  if c cmpeq oo then return (d-b)/(d-a); end if;
  if d cmpeq oo then return (c-a)/(c-b); end if;
  return (a-c)*(b-d)/(b-c)/(a-d);
catch e
  return oo;
end try;
end function;

function mysqrt(x, plc)
  K:= myNF(plc);
  p:= Polynomial([K | -x,0, 1]);
  r:= Roots(p);
  if #r ne 0 then return myEval(r[1,1], plc) ge 0 select r[1,1] else r[2,1], plc; end if;
  L:= ext< K | p >;
  D:= [ d[1] : d in Decomposition(L, plc) ];
  return L.1, myEval(L.1, D[1]) ge 0 select D[1] else D[2];
end function;

apply:= function(X, x)
  if x cmpeq oo then
    return X[2,1] eq 0 select oo else X[1,1]/X[2,1];
  end if;
  d:= X[2,1] * x + X[2,2];
  return d eq 0 select oo else (X[1,1] * x + X[1,2])/d;
end function;

fixed_exact:= function(X, plc)
  a,b,c,d:= Explode(Eltseq(X));
  disc:= (a+d)^2-4;
  if disc eq 0 then
    return (c eq 0 select [oo] else [(a-d) / (2*c)]), plc;
  end if;
  if myEval(myNF(plc) ! disc, plc) lt 0 then return [], plc; end if;
  if c eq 0 then return [* -b/(a-d), oo *], plc; end if;
  D, plc:= mysqrt(disc, plc); c:= 2*c;
  x:= a-d;
  return Sort([(x-D)/c, (x+D)/c], func< r,s | myEval(r-s, plc) >), plc;
end function;

// Diagonalize an hyperbolic element. The larger eigenvalue comes first.
function diagonalize(A, plc)
  if IsDiagonal(A) then
    S:= Parent(A) ! 1;
  elif A[2,1] eq 0 then
    S:= Matrix(2, [(A[1,1] - A[2,2]) / A[1,2], 1, 0, 1]);
    A[1,2]:= 0;
  else
    f, plc:= fixed_exact(A, plc);
    assert #f eq 2;
    S:= Matrix(2, [1, -f[1], -1, f[2]]);  // Det > 0
    A:= S*A*S^-1;
    assert IsDiagonal(A);
  end if;
  if myEval(A[1,1] - A[2,2], plc) ge 0 then return A, S, plc; end if;
  return DiagonalMatrix([A[2,2], A[1,1]]), Matrix(2, [BaseRing(A) | 0, 1, -1, 0]) * S, plc;
end function;

eqn:= function(z,w)
  if z cmpeq oo then
    return w cmpeq oo select [0, 0, 1, 0] else  [1, 0, -w, 0];
  end if;
  return w cmpeq oo select [0,0,z,1] else [z, 1, -w*z, -w];
end function;

function Transformation(Points)
  M:= Matrix([ eqn(P[1], P[2]) : P in Points ]);
  K:= Kernel(Transpose(M));
  assert Dimension(K) eq 1;
  return Matrix(2, Eltseq(K.1));
end function;

// decompose X into z:-> b*z + a
function get_cyclic_affine(X)
  a:= apply(X, 0);
  return a, apply(X, 1) - a;
end function;

procedure MarkElementary(G)
  A:= G`A; B:= G`B; plc:= G`plc;
  G`type:= "nd";
  f, plc:= fixed_exact(A, plc);
  same:= [* x: x in f | apply(B, x) cmpeq x *];
  if #same eq 1 then
    if same[1] cmpne oo then
      X:= Exclude([myNF(plc) | 0,1,-1], same[1]);
      S:= Transformation( [* <same[1], oo> *] cat [* <X[1], 0>, <X[2], 1> *] );
      SI:= S^-1; A:= S*A*SI; B:= S*B*SI; 
    end if;
    assert A[2,1] eq 0 and B[2,1] eq 0;
    _, a:= get_cyclic_affine(A);
    _, b:= get_cyclic_affine(B);
    if IsOne(a^12) and IsOne(b^12) then
      Orbit:= {@ 1 @};
      Path:= [ A^0 ];
      Stab:= {};
      i:= 0;
      while i lt #Orbit do
        i +:= 1;
        x:= a*Orbit[i]; idx:= Index(Orbit, x); if idx eq 0 then Include(~Orbit, x); Append(~Path, A*Path[i]); else Include(~Stab, Path[idx]^-1*A*Path[i]); end if;
        x:= b*Orbit[i]; idx:= Index(Orbit, x); if idx eq 0 then Include(~Orbit, x); Append(~Path, B*Path[i]); else Include(~Stab, Path[idx]^-1*B*Path[i]); end if;
      end while;
      Gens:= [];
      for s in Stab do
        x, y:= get_cyclic_affine(s);
	assert y eq 1; Append(~Gens, x);
      end for;
      L:= Lattice( Matrix( Rationals(), [ Type(y) eq FldRatElt select [y] else Flat(y) : y in Gens ]) );
      if Rank(L) le 2 then G`type:= "ed"; end if;
    end if;
  end if;
end procedure;




// The user interface
intrinsic Print(G::GrpSL2Gen)
{internal}
  printf "Two generator subgroup of %oSL(2,R) generated by\n%o,\n%o", G`PSL select "P" else "", G`A, G`B;
end intrinsic;

intrinsic IsCoercible(G::GrpSL2Gen, A::. ) -> BoolElt, .
{internal}
  return false, "invalid coercion";
end intrinsic;

function create(A,B,plc,PSL)
  T:= Type(plc);
  if T ne Infty and not (ISA(T, PlcNumElt) and IsReal(plc)) then
    return false, "The third argument must be a real infinite place";
  elif Determinant(A) ne 1 or Determinant(B) ne 1 then
    return false, "The matrices must have determinant 1.";
  end if;
  K:= myNF(plc);
  okA, A:= CanChangeRing(A, K);
  okB, B:= CanChangeRing(B, K);
  if not (okA and okB) then return false, "The base rings of A and B must be the base field of the place."; end if;
  G:= New(GrpSL2Gen);
  G`A:= A; G`B:= B; G`K:= K; G`plc:= plc; G`PSL:= PSL;
  return true, G;
end function;

intrinsic TwoGeneratorSubgroupSL2R(A::Mtrx, B::Mtrx, plc::.) -> GrpSL2Gen
{The subgroup of SL(2,R) generated by A, B using the embedding plc}
  ok, G:= create(A,B,plc,false);
  require ok : G;
  return G;
end intrinsic;

intrinsic TwoGeneratorSubgroupPSL2R(A::Mtrx, B::Mtrx, plc::.) -> GrpSL2Gen
{The subgroup of SL(2,R) generated by A, B using the embedding plc}
  ok, G:= create(A,B,plc,true);
  require ok : G;
  return G;
end intrinsic;

intrinsic '.'(G::GrpSL2Gen, i::RngIntElt) -> Mtrx
{The i-th generator of the group G}
  require i in [-2..2] : "Index out of range";
  return (Abs(i) eq 2 select G`B else G`A)^Sign(i);
end intrinsic;

// set up G`type and U, V if necessary
procedure prepare(G)
  if assigned G`type then return; end if;

  plc:= G`plc;
  U:= normalize(G`A, plc);
  V:= normalize(G`B, plc);
  TU:= Trace(U); TV:= Trace(V);
  if (TU eq 2 or Evaluate(TU, plc) le 2) and (TV eq 2 or Evaluate(TV, plc) le 2) and IsFinite(G) then
    G`type:= "finite";
    return;
  end if; 

  T:= Trace(U*V*U^-1*V^-1);
  if T eq 2 then MarkElementary(G); return; end if;

  F:= SLPGroup(2);
  WU:= F.1; WV:= F.2;

  if myEval(T, G`plc) le -2 or T eq -2 then 
    G`type:= <0,0,0>; 
  else
    if (myEval(TU, plc) lt 2 and TU ne 2) or (myEval(TV, plc) lt 2 and TV ne 2) then
      G`type:= "not free";
      return;
    end if;
    done:= false;
    if myEval(TV - TU, plc) lt 0 then swap(~U, ~V); swap(~TU, ~TV); swap(~WU, ~WV); end if; 
    repeat
      S:= [ U*V, U*V^-1 ];
      T:= [ Trace(s): s in S ];
      E:= [ myEval(t, plc): t in T ];
      A:= [ Abs(e): e in E ];
      m, i:= Minimum(A);
      if m lt 2 and T[i] notin {-2,2} then 
        G`type:= "not free";
	return;
      end if;
      s:= Sign(E[i]);
      if m lt myEval(TU, plc) then 
        V:= U; TV:= TU; U:= s*S[i]; TU:= s*T[i]; tmp:= WV; WV:= WU; WU:= WU*tmp^(i eq 1 select 1 else -1);
      elif m lt myEval(TV, plc) then
        V:= s*S[i]; TV:= s*T[i]; WV:= WU*WV^(i eq 1 select 1 else -1);
      else
        done:= true;
      end if;
    until done;
    if myEval(Trace(U*V^-1), plc) ge 0 then V:= V^-1; WV:= WV^-1; end if;
    // TODO: change this for testing discreteness
    G`type:= <0,0,0>;
  end if;
  G`U:= U; G`V:= V; G`WordsUV:= [WU, WV];
end procedure;

intrinsic IsFinite(G::GrpSL2Gen) -> BoolElt
{Returns true if and only if G is discrete and free}
  if not assigned G`finite then
    GG:= sub< GL(2, G`K) | G`A, G`B >;
    if IsFinite(GG) then
      G`finite:= <true, GG>;
    else
      G`finite:= <false>;
    end if;
  end if;
  return G`finite[1];
end intrinsic;


// test if G is free and initializes U,V if necessary
intrinsic IsDiscreteAndFree(G::GrpSL2Gen) -> BoolElt
{Returns true if and only if G is discrete and free}
  prepare(G);
  return G`type cmpeq <0,0,0>;
end intrinsic;

function std_form(A, plc)
  a:= A[1,1];
  assert A[2,2] eq a;
  s, plc:= mysqrt(a^2-1, plc);
  S:= Matrix(2, [1, x, 1, -x]) where x:= s/A[2,1];
  return S, (a+s)^2, plc;
end function;

function get_exponent1(z, c, xy, r, plc: S:= false, Tries:= 3)
  zz:= S cmpeq false select z else apply(S[1], z);
  l:= l2norm(zz);
  if l-2*z[1]*c+xy eq 0 then return 0, z; end if;	// we are on the boundary
  czl:= c*zz[1]/l;
  i:= 0;
  while i lt Tries do
    prec:= 300 + 100*i;
    try
      tmp:= myEval(czl, plc: ForceReal, Precision:= prec) + Sqrt( myEval(czl^2 - xy/l, plc: ForceReal, Precision:= prec) );
      tmp:= Log(tmp) / Log(myEval(r, plc : ForceReal, Precision:= prec));
      k:= Ceiling(tmp);
      zz *:= r^k;
      if S cmpne false then zz:= apply(S[2], zz); end if;
      return k, zz;
    catch e ;
    end try;
    i+:= 1;
  end while;
  return 0, z;
end function;

// Solves membership in case 1 for PSL.
function Membership1(G, M)
//  "entering 1";
  if not assigned G`S then
    _, S, G`plc:= diagonalize(G`U, G`plc);
    G`S:= DiagonalMatrix([(S*G`V*S^-1)[2,1], 1]) * S;
  end if;
  plc:= G`plc; S:= G`S; SI:= S^-1;
  U:= S * G`U * SI; V:= S * G`V * SI; VI:= V^-1;

  // construct domain and init
  a,b,_,d:= Explode(Eltseq(V)); r:= U[1,1]^2;
  R2P:= a/d; R2M:= d/a;
  rR:= myEval(r, plc: ForceReal);
  s, plc1:= mysqrt(R2P, plc);

  if Trace(G`A*G`B*G`A^-1*G`B^-1) eq -2 then
    x:= a-s; y:= -d+1/s;
    assert plc eq plc1;
  else // We can get away without using s. This usually safes one relative extension.
    sR:= myEval(s, plc1);
    x:= RationalReconstruction( (aR + sR)/rR, aR - sR ) where aR:= myEval(a, plc);
    y:= RationalReconstruction( dR + 1/sR, (dR - 1/sR)/rR ) where dR:= myEval(-d, plc);
  end if;
  xy:= x*y;

  VV, SV, plc1:= diagonalize(V, plc1);
  SVI:= SV^-1;
  kk:= VV[1,1]^2;
  xV:= apply(SV, -d+1/s); yV:=apply(SV, -d-1/s); cV:= (xV+yV)/2; xVyV:= xV*yV;

  // the inner circle is centered at (c,  0) with radius r1
  // the outer circle is centered at (r*c,0) with radius r*r1
  c:= (x+y)/2; r1:= (x-y)/2;

  L := ext< myNF(plc ) | Polynomial([1,0,1]) >;           // plc  is real, so this works.
  L1:= ext< myNF(plc1) | Polynomial([1,0,1]) >;           // plc1 is real, so this works.
  t:= L.1 * RationalReconstruction(s, rR*s) where s:= Sqrt(myEval(-xy, plc));
  assert InCircle(t, r*c, (r*r1)^2, plc) and not InCircle(t, c, r1^2, plc);

  // start algorithm
  z:= apply(S*M*SI, t);
  F:= FreeGroup(2);
  w:= F ! 1;

  repeat
    // get power of U
    k, z:= get_exponent1(z, c, xy, r, plc: Tries:= oo);
    w:= w*F.1^-k;
    //"applied U^", -k;

    k:= 0;
    while InCircle(z, -d, R2M, plc) and  k le MAXTRIES do k+:= 1; z:= apply(V,  z); end while;
    while InCircle(z,  a, R2P, plc) and -k le MAXTRIES do k-:= 1; z:= apply(VI, z); end while;
    if Abs(k) gt MAXTRIES then
      kV, zz:= get_exponent1(L1 ! Eltseq(z), cV, xVyV, kk, plc1: S:= [SV, SVI]);
      z:= L ! Eltseq(zz);
      k +:= kV;
    end if;
    w:= w*F.2^-k;
    //"applied V^", -k;
  until k eq 0;

  return z eq t, w;
end function;

function get_exponent2(z, r, S, SI, plc)
  zz:= apply(S, z);
  l:= l2norm(zz);
  if l eq r or l eq 1/r then return 0, z; end if;
  i:= 0;
  while i lt 5 do
    prec:= 300 + 100*i;
    try
      k:= Floor((1 - Log(myEval(l, plc : Precision:= prec, ForceReal)) / Log(myEval(r, plc : Precision:= prec, ForceReal)))/2);
      return k, apply(SI, zz * r^k);
    catch e ;
    end try;
    i+:= 1;
  end while;
  return 0, z;
end function;

// Solves membership in case 2 for PSL.
function Membership2(G, M)
// "entering 2";
  tU:= Trace(G`U) eq 2; tV:= Trace(G`V) eq 2;
  assert tU or not tV;          // tV implies tU by choice of generators.

  // We need to find a conjugating matrix than makes U,V "nice".
  if not assigned G`S then
    plc:= G`plc;
    // First, normalize the fix points of U, V.
    f, plc:= fixed_exact(G`U, plc);
    Points:= tU select [* <f[1], oo> *] else [* <f[1], -1>, <f[2], 1> *];
    g, plc:= fixed_exact(G`V, plc);
    if tV then
      Append(~Points, <g[1], 0>);
      ok:= exists(t){i: i in [1,-1,0] | f[1] cmpne i and g[1] cmpne i}; assert ok;
      Append(~Points, <t, 1>);
    else
      if tU then k:= 1;
      else
        s, plc:= mysqrt(cross(f[1], f[2], g[1], g[2]), plc);
        k:= (1+s)/(1-s);
      end if;
      Append(~Points, <g[1], -k>);
      Append(~Points, <g[2],  k>);
    end if;
    S:= Transformation( Points );
    if myEval((S*G`U*S^-1)[1,2], plc) lt 0 then
      S:= DiagonalMatrix([1,-1]) * S;
    end if;
    G`plc:= plc; G`S:= S;
  end if;

  plc:= G`plc; S:= G`S; SI:= S^-1;
  U:= S*G`U*SI; V:= S*G`V*SI;
  assert myEval(V[2,1], plc) gt 0;

  // Now the reconstruction begins
  if tU then
    n:= U[1,2]; n2:= n/2;
  else
    cU:= U[1,1] / U[2,1]; rU:= 1/U[2,1]^2; UI:= U^-1;
    SU, rU2, plcU:= std_form(U, plc); SUI:= SU^-1;
  end if;
  if tV then
    m:= V[2,1]; m2:= m/2;
  else
    cV:= V[1,1] / V[2,1]; rV:= 1/V[2,1]^2; VI:= V^-1;
    SV, rV2, plcV:= std_form(V, plc); SVI:= SV^-1;
  end if;

  L:= ext< myNF(plc) | Polynomial([1,0,1]) >;
  if not tU then LU:= ext< myNF(plcU) | Polynomial([1,0,1]) >; end if;
  if not tV then LV:= ext< myNF(plcV) | Polynomial([1,0,1]) >; end if;
  z:= apply(S * M * SI, L.1);
  F:= FreeGroup(2);
  w:= F ! 1;

  eps:= 10^-10;

  repeat
    if tU then
      e:= myEval(z[1]/n, plc); 
      // are we sitting on the boundary?
      if Abs(Abs(e)-0.5) lt eps and (z[1] eq n2 or z[1] eq -n2) then break; end if;
      k:= -Round(e);
      z+:= k*n;
    else
      k:= 0;
      while InCircle(z, -cU, rU, plc) and  k le MAXTRIES do k+:= 1; z:= apply(U,  z); end while;
      while InCircle(z,  cU, rU, plc) and -k le MAXTRIES do k-:= 1; z:= apply(UI, z); end while;
      if Abs(k) gt MAXTRIES then
        kU, z:= get_exponent2(LU ! Eltseq(z), rU2, SU, SUI, plc);
        k +:= kU;
        z:= L ! Eltseq(z);
      end if;
     end if;
    w:= w*F.1^-k;
//    "U", -k;

    if tV then
      // Transform the plane via [0,1,-1,0]. Then V looks like U and the task becomes trivial.
      zz:= 1/z;
      e:= myEval(zz[1]/m, plc);
      if Abs(Abs(e)-0.5) lt eps and (zz[1] eq m2 or zz[1] eq -m2) then break; end if;
      k:= -Round(e);
      z:= 1/(zz+k*m);
    else
      k:= 0;
      while InCircle(z, -cV, rV, plc) and  k le MAXTRIES do k+:= 1; z:= apply(V,  z); end while;
      while InCircle(z,  cV, rV, plc) and -k le MAXTRIES do k-:= 1; z:= apply(VI, z); end while;
      if Abs(k) gt MAXTRIES then
        kV, z:= get_exponent2(LV ! Eltseq(z), rV2, SV, SVI, plc);
        k +:= kV;
        z:= L ! Eltseq(z);
      end if;
    end if;
    w:= w*F.2^-k;
//     "V", -k;
  until k eq 0;

  return z eq L.1, w;
end function;

intrinsic Membership(G::GrpSL2Gen, M::Mtrx) -> BoolElt, GrpFPElt
{Returns true if M is in G. If this is the case, a word expressing M as the generators of G is also returned}
  require Ncols(M) eq Nrows(M) and Determinant(M) eq 1 : "The matrix must have determinant 1";
  ok, M:= CanChangeRing(M, G`K);
  require ok : "The group and the matrix must be over the same field.";

  if IsFinite(G) then
    GG:= G`finite[2];
    ok, M:= IsCoercible(GG, M);
    if not ok and G`PSL then 
      ok, M:= IsCoercible(GG, -M);
    end if;
    if not ok then return false, _; end if;
    _, h:= WordGroup(GG);
    return true, Evaluate(M @@ h, FreeGroup(2));
  end if;

  // Do NOT remove the following line. It fills the data structure if necessary!
  require IsDiscreteAndFree(G) : "The group is not discrete and free.";

  // First, solve membership in \PSL.
  t:= Trace(G`A*G`B*G`A^-1*G`B^-1);
  if myEval(t+2, G`plc) le 0 or t eq -2 then
    ok, w:= Membership1(G, M);
  else
    ok, w:= Membership2(G, M);
  end if;
  if ok then 
    // w is the only possible candidate (in U,V !)
    // Rewrite w as a word in A,B
    F:= Parent(w);
    if assigned G`WordsUV then
      w:= w @ h where h:= hom< F -> F | Evaluate(G`WordsUV[1], F), Evaluate(G`WordsUV[2], F) >;
    end if;
    // Test if w does the trick
    N:= w @ h where h:= hom < F -> GL(2, G`K) | G`A, G`B >;
    if N eq M or (G`PSL and N eq -M) then return true, w; end if;
  end if;
  return false, _;
end intrinsic;

intrinsic WitnessPair(G::GrpSL2Gen)-> Mtrx, Mtrx, []
{Returns a witness pair for G as well as SLPs expressing the pair in terms of the generators}
  prepare(G);
  require assigned G`U : "The group has no witness pair";
  return G`U, G`V, G`WordsUV;
end intrinsic;

